classdef ooGmsh2 < handle  
% CLASSDEF ooGmsh2:
%   
% works with gmsh 2.16.0 to 3.0.6. with versions>=4.0.0 no negative tag for interfaces???
%
% Copyright (C) 2017-2020 F. Cuvelier
  properties
    q=[];  % Nodes/Vertices
    dim=3; % Space dimension
    nq=0;  % number of Nodes
    d=0;   % max(EntityDim) or d-simplicial mesh
    % Ms=[]; % Splitted elements by elmt_type OLD
    sElts=[]; % Splitted elements by elmt_type
    toGlobal=[];
    partitionnedfile=false; % 
    orders = []; % list of orders of the elements
    types =[];   % list of types
    %debug={};
    MeshFormat=[];
    meshfile=[];
    Info=[];
    debug=[];
  end
  
  methods
    function self=ooGmsh2(gmsh_file)
    % ooGmsh2 constructor, create an ooGmsh2 object from a gmsh mesh file.
    %   
    % USAGE:
    %   Gh=ooGmsh2(gmsh_file);
    %
    % INPUT:
    %   gmsh_file: name (string) of the gmsh mesh file to read.
    %
    % OUTPUT:
    %   Gh: an ooGmsh2 object.
    %
    % Copyright (C) 2017-2020 F. Cuvelier
      [fid,message]=fopen(gmsh_file,'r');
      if ( fid == -1 )
          error([message,' : ',gmsh_file]);
      end
      self.meshfile=gmsh_file;
      self.MeshFormat=fc_oogmsh.msh.read_MeshFormat(fid);
      assert( strcmp(self.MeshFormat.version,'2.2') , 'Needs version 2.2 of msh file format! Found version %s',self.MeshFormat.version )
      while ( 1 )
        text = fgetl ( fid );
        if ( text == -1 )
          error(['string $Nodes not found in file ',gmsh_file])
        end
        if strcmp(strtok(text),'$Nodes')
          break
        end
      end
      nq=fscanf(fid,'%d',1);
      R=fscanf(fid,'%d %f %f %f',[4,nq]);
      self.toGlobal=R(1,:);
      %if (min(self.toGlobal)~=1) || (max(self.toGlobal)~=nq), self.partitionnedfile=true;end
      self.q=R(2:end,:);
      if sum(abs(self.q(3,:)))==0,self.q=self.q(1:2,:);self.dim=2;end
      if sum(abs(self.q(2,:)))==0,self.q=self.q(1,:);self.dim=1;end
      self.nq=size(self.q,2);
      while ( 1 )
        text = fgetl ( fid );
        if ( text == -1 )
          error(['string $Elements not found in file ',gmsh_file])
        end
        if strcmp(strtok(text),'$Elements')
          break
        end
      end
      self.splitByType(fid);
      self.orders=unique(cell2mat({self.sElts(:).order}));
      self.types=cell2mat({self.sElts(:).type});
      for i=1:length(self.sElts)
        if ~isempty(self.sElts(i).part_lab), self.partitionnedfile=true;end
      end
      self.Info.meshfile=gmsh_file;
      self.set_d();
    end % ooGmsh2 constructor
    
    function disp(self)
      fc_tools.comp.dispobject(self)
    end
    
    function info(self,varargin)
      fc_tools.comp.infoobject(self,varargin{:})
    end
    
    function p2g=phy2geo(self)
      ne=length(self.sElts);
      p2g=cell(1,ne);
      for i=1:ne
        phylab=unique(self.sElts(i).phys_lab)';
        npl=length(phylab);
        Geos=cell(1,npl);
        geolab=[];Phys=[];
        for l=1:npl
          Geos{l}=unique(self.sElts(i).geo_lab((self.sElts(i).phys_lab==phylab(l))))';
          geolab=[geolab, Geos{l}];
          Phys=[Phys,phylab(l)*ones(1,length(Geos{l}))];
        end
        p2g{i}=struct('type',self.sElts(i).type,'phys_labs',phylab,'toGeo',{Geos},'geo_labs',geolab,'toPhys',Phys);
      end
    end
    
    function self=splitByType(self,fid)
    % FUNCTION splitByType, method function of the ooGmsh class.
    %   Internal. Must be private in future release.
    %
    % Copyright (C) 2017-2020 F. Cuvelier

    % elm-number elm-type number-of-tags < tag > ... node-number-list
    %  number-of-tags :
    %    gives the number of integer tags that follow for the n-th element. By default,
    %    the first tag is the number of the physical entity to which the element belongs;
    %    the second is the number of the elementary geometrical entity to which the
    %    element belongs; the third is the number of mesh partitions to which the element
    %    belongs, followed by the partition ids (negative partition ids indicate ghost
    %    cells). A zero tag is equivalent to no tag. Gmsh and most codes using the MSH
    %    2 format require at least the first two tags (physical and elementary tags).
      elt_struct=@(d,me,nme,type,order,geo,PhysicalLabel,GeometricalLabel,nMeshPart,MeshPart,nTags) ...
          struct('type',type,'geo',geo,'d',d,'order',order,'me',me,'nme',nme,'phys_lab',PhysicalLabel, ...
                'geo_lab',GeometricalLabel,'part_lab',MeshPart,'nb_parts',nMeshPart);
      elm_types=fc_oogmsh.gmsh.elm_type_list();
      Mt=self.getElements(fid); % Get All elements
      self.debug.Mt=Mt;
      
      index=struct('num',1,'elm_type',2,'nb_tags',3,'phys_lab',4,'geo_lab',5,'nb_parts',6);
      Ltype=unique(Mt(:,index.elm_type)); % gets all elm_type of the meshes
      self.sElts=cell(length(Ltype),1); % Will converted in struct array 
      
      for i=1:length(Ltype)
        I=find(Mt(:,index.elm_type)==Ltype(i));
        M=double(Mt(I,:));
        type=Ltype(i);
        order=elm_types{type}.order;
        geo=elm_types{type}.geo;
        d=elm_types{type}.d;
        nme=length(I);
        nTags=M(:,index.nb_tags)-3; % Tags number other that 'phys_lab' and 'geo_lab', if num<=0 => no tags
        PhysicalLabel=M(:,index.phys_lab);
        GeometricalLabel=M(:,index.geo_lab);
        nMeshPart=zeros(nme,1); % number of mesh partitions to which the element belongs
        MeshPart=[];  
        K=find(nTags>=0); % index of elmts with partition ids
        if ~isempty(K), 
          Kc=setdiff(1:nme,K); % index of elmts without partition ids
          nMeshPart(K)=M(K,index.nb_parts);
          max(nTags);
          MeshPart=zeros(nme,max(nTags)); % Don't store  'phys_lab','geo_lab' and 'nb_parts'
        else
          Kc=1:nme;
        end
        ndfe=elm_types{type}.nb_nodes;
        me=zeros(ndfe,nme);
        % read elmts without partition ids
        if ~isempty(Kc)
          me(:,Kc)=M(Kc,index.geo_lab+[1:ndfe])';
        end
        
        % read elmts with partition ids
        K=K';
        lnTags=unique(nTags(K)); % list of number of partitions 
        
        for j=1:numel(lnTags)
          nt=lnTags(j);
          ik=find(nTags(K)==nt);
          k=K(ik);
          MeshPart(k,1:nt)=M(k,index.nb_parts+[1:nt]);
          me(:,k)=M(k,index.nb_parts+nt+[1:ndfe])';
        end
        self.sElts{i}=elt_struct(d,me,nme,type,order,geo,PhysicalLabel,GeometricalLabel,nMeshPart,MeshPart,nTags);
      end
      self.sElts=cell2mat(self.sElts);
    end

    function M=getElements(self,fid)
    % FUNCTION getElements, method function of the ooGmsh class.
    %   Internal. Must be private in future release.
    %
    % Copyright (C) 2017-2020 F. Cuvelier
      nelem=fscanf(fid,'%d',1);
      if fc_tools.comp.isOctave()
        M=zeros(nelem,55); % 55 a ameliorer
        fgetl(fid);
        for i=1:nelem
          A=sscanf(fgetl(fid),'%d')';
          %nM(i)=numel(A);
          M(i,1:numel(A))=A;
        end
      else
        format = repmat('%d',1,55); % 55 a ameliorer
        R = textscan(fid,format,nelem);
        M = [R{:}];
      end
    end
    
    function [q,me]=extractElement(self,type,varargin)
      it=find(self.types==type);
      assert(length(it)==1)
      p = inputParser;
      p.KeepUnmatched=true; 
      p.PartialMatching=false;
      p.addParamValue('geo_labs',[]); 
      p.addParamValue('phys_labs',[]);
      p.parse(varargin{:});
      R=p.Results;
      if ~isempty(R.geo_labs) && ~isempty(R.phys_labs)
        error('Unallowed used of options')
      end
      if isempty(R.geo_labs) && isempty(R.phys_labs)
        [idx,il,me]=unique(self.sElts(it).me(:));
        q=self.q(:,idx);
        me=reshape(me,[self.sElts(it).d+1,self.sElts(it).nme]);
        return
      end
      if ~isempty(R.geo_labs)
        labs=self.sElts(it).geo_lab;
        olabs=R.geo_labs;
      else
        labs=self.sElts(it).phys_lab;
        olabs=R.phys_labs;
      end
      ME=[];
      for i=1:length(olabs)
        lab=olabs(i);
        ME=[ME,self.sElts(it).me(:,labs==lab)];
      end
      if isempty(ME)
        q=[];me=[];return
      end
      [idx,il,me]=unique(ME(:));
      q=self.q(:,idx);
      me=reshape(me,[self.sElts(it).d+1,size(ME,2)]);
    end
    
    function [q,tags]=get_q(self,varargin)
      q=self.q;
      tags=self.toGlobal;
    end
    
    function set_d(self)
      self.d=max([self.sElts(:).d]);
    end
    
    function sElts=get_sElts_by_entityDim(self,entityDim)
      k=find([self.sElts(:).d]==entityDim);
      if isempty(k), sElts=[]; else, sElts=self.sElts(k);end
    end
    
    function tags=get_PhysicalTags(self,type)
      if ~ismember(type,self.types), tags=[];return;end
      idx=find(self.types==type);
      tags=unique([self.sElts(idx).phys_lab])';
    end 
    
    function tags=get_PhysicalTags_entityDim(self,entityDim)
      sElts=get_sElts_by_entityDim(self,entityDim);
      tags=unique(sElts.phys_lab)';
    end 
    
    function tags=get_ElementaryTags(self,type)
      if ~ismember(type,self.types), tags=[];return;end
      idx=find(self.types==type);
      tags=unique(self.sElts(idx).geo_lab)';
    end 
    
    function me=get_me_PhysicalTag(self,type,phtag)
      if ~ismember(type,self.types), me=[];return;end
      i=find(self.types==type);
      idx=find(self.sElts(i).phys_lab==phtag);
      me=self.sElts(i).me(:,idx);
    end
    
    function me=get_me_ElementaryTag(self,type,eltag)
      if ~ismember(type,self.types), tags=[];return;end
      %assert(ismember(type,self.types))
      i=find(self.types==type);
      idx=find(self.sElts(i).geo_lab==eltag);
      me=self.sElts(i).me(:,idx);
    end
    
    function varargout=get_localmesh_PhysicalTag(self,type,PhysTag)
      assert( ismember(nargout,[2,3]) )
      assert(ismember(type,self.types))
      % returns  [q,me] or [q,me,toGlobal]
      me=self.get_me_PhysicalTag(type,PhysTag);
      [q,qtags]=self.get_q();
      [q,me,toGlobal]=fc_oogmsh.msh.FromGlobalToLocal(q,me,qtags);
      varargout{1}=q;
      varargout{2}=me;
      if nargout==3,varargout{3}=toGlobal;end  
    end  
    
    function varargout=get_localmesh_ElementaryTag(self,type,EltTag)
      assert( ismember(nargout,[2,3]) )
      assert(ismember(type,self.types))
      % returns  [q,me] or [q,me,toGlobal]
      me=self.get_me_ElementaryTag(type,EltTag);
      [q,qtags]=self.get_q();
      [q,me,toGlobal]=fc_oogmsh.msh.FromGlobalToLocal(q,me,qtags);
      varargout{1}=q;
      varargout{2}=me;
      if nargout==3,varargout{3}=toGlobal;end  
    end  
    
    function [me,mel]=extractInterfaceElements(self,d)
      me=[];mel=[];
      k=find(cell2mat({self.sElts(:).d})==d);
      if isempty(k),return;end
      I=find(self.sElts(k).geo_lab<0);
      if isempty(I),return;end
      me=self.sElts(k).me(:,I);
      mel=self.sElts(k).geo_lab(I);
    end
    
    function plotElementaryElements(Gh,varargin)
      fc_oogmsh.graphics.ooGmsh2.plotElementaryElements(Gh,varargin{:})
    end
    
    function plotInterfaceElements(Gh,varargin)
      fc_oogmsh.graphics.ooGmsh2.plotInterfaceElements(Gh,varargin{:})
    end
    
    function plotPartitionElements(Gh,varargin)
      fc_oogmsh.graphics.ooGmsh2.plotPartitionElements(Gh,varargin{:})
    end
    
    function plotPhysicalElements(Gh,varargin)
      fc_oogmsh.graphics.ooGmsh2.plotPhysicalElements(Gh,varargin{:})
    end
    
    function plotmesh(Gh,d,varargin)
      fc_oogmsh.graphics.ooGmsh2.plotmesh(Gh,d,varargin{:})
    end
  end % methods
end %

